import numpy as np
import pandas as pd
from sklearn.preprocessing import MinMaxScaler, StandardScaler
from sklearn.model_selection import train_test_split
import matplotlib.pyplot as plt
import h5py
import scipy.misc as im
import os, pickle


class Dataset:
    
    def __init__(self, config, logs_dir, data_dir, dataset_name, split_ratio=0.2):
        self.input_shape = config[dataset_name]["model"]["input_shape"]
        self.image_shape = config[dataset_name]["image_size"]
        self.logs_dir = logs_dir
        self.data_dir = data_dir
        
        if dataset_name.lower() == "mnist" or dataset_name.lower() == "fashion_mnist":
            from keras.datasets import mnist, fashion_mnist
            from tensorflow import keras
            
            if dataset_name.lower() == "mnist":
                dataset = keras.datasets.mnist
                self.class_names = ['Zero', 'One', 'Two', 'Three', 'Four',
                                    'Five', 'Six', 'Seven', 'Eight', 'Nine']
            elif dataset_name.lower() == "fashion_mnist":
                dataset = keras.datasets.fashion_mnist
                self.class_names = ['T-shirt/top', 'Trouser', 'Pullover', 'Dress', 'Coat',
                                    'Sandal', 'Shirt', 'Sneaker', 'Bag', 'Ankle boot']
            
            (self.train_images, self.train_labels), (self.test_images, self.test_labels) = dataset.load_data()
            
            #### Split dataset into labeled and unlabeled ####
            self.train_images = np.array([x[1] for x in sorted(zip(self.train_labels, self.train_images), key=lambda tup: tup[0])])
            self.train_labels = np.array([x[0] for x in sorted(zip(self.train_labels, self.train_images), key=lambda tup: tup[0])])
            self.test_images = np.array([x[1] for x in sorted(zip(self.test_labels, self.test_images), key=lambda tup: tup[0])])
            self.test_labels = np.array([x[0] for x in sorted(zip(self.test_labels, self.test_images), key=lambda tup: tup[0])])
            
            train_images, train_labels, test_images, test_labels, remaining_images, remaining_labels = [], [], [], [], [], []
            start = 0
            for i in range(len(self.class_names)):
                class_label = i
                class_len = len(np.where(self.train_labels == class_label)[0])
                class_split_len = int(class_len * split_ratio)
                print(class_len, class_split_len, start+class_split_len)
                train_images.extend(self.train_images[start:start+class_split_len])
                train_labels.extend(self.train_labels[start:start+class_split_len])
                remaining_images.extend(self.train_images[start+class_split_len:start+class_len])
                remaining_labels.extend(self.train_labels[start+class_split_len:start+class_len])
                start += class_len
            
            remaining_images.extend(self.test_images)
            remaining_labels.extend(self.test_labels)
            
            start = 0
            for i in range(len(self.class_names)):
                class_label = i
                class_len = len(np.where(self.test_labels == class_label)[0])
                class_split_len = int(class_len * split_ratio)
                print(class_len, class_split_len)
                test_images.extend(self.test_images[start:start+class_split_len])
                test_labels.extend(self.test_labels[start:start+class_split_len])
                start += class_len
            
            print(len(train_images), len(train_labels), len(test_images), len(test_labels))
            self.train_images, self.train_labels = np.array(train_images), np.array(train_labels)
            self.test_images, self.test_labels = np.array(test_images), np.array(test_labels)
            self.remaining_images, self.remaining_labels = np.array(remaining_images), np.array(remaining_labels)
                
            ##################################################
            
        elif dataset_name.lower() == "usps":
            with h5py.File(data_dir + 'usps.h5', 'r') as hf:
                    train = hf.get('train')
                    self.train_images = train.get('data')[:]
                    self.train_labels = train.get('target')[:]
                    test = hf.get('test')
                    self.test_images = test.get('data')[:]
                    self.test_labels = test.get('target')[:]
                    self.class_names = ['Zero', 'One', 'Two', 'Three', 'Four',
                                        'Five', 'Six', 'Seven', 'Eight', 'Nine']
            #### Split dataset into labeled and unlabeled ####
            self.train_images = np.array([x[1] for x in sorted(zip(self.train_labels, self.train_images), key=lambda tup: tup[0])])
            self.train_labels = np.array([x[0] for x in sorted(zip(self.train_labels, self.train_images), key=lambda tup: tup[0])])
            self.test_images = np.array([x[1] for x in sorted(zip(self.test_labels, self.test_images), key=lambda tup: tup[0])])
            self.test_labels = np.array([x[0] for x in sorted(zip(self.test_labels, self.test_images), key=lambda tup: tup[0])])
            
            train_images, train_labels, test_images, test_labels, remaining_images, remaining_labels = [], [], [], [], [], []
            start = 0
            for i in range(len(self.class_names)):
                class_label = i
                class_len = len(np.where(self.train_labels == class_label)[0])
                class_split_len = int(class_len * split_ratio)
                print(class_len, class_split_len, start+class_split_len)
                train_images.extend(self.train_images[start:start+class_split_len])
                train_labels.extend(self.train_labels[start:start+class_split_len])
                remaining_images.extend(self.train_images[start+class_split_len:start+class_len])
                remaining_labels.extend(self.train_labels[start+class_split_len:start+class_len])
                start += class_len
            
            remaining_images.extend(self.test_images)
            remaining_labels.extend(self.test_labels)
            
            start = 0
            for i in range(len(self.class_names)):
                class_label = i
                class_len = len(np.where(self.test_labels == class_label)[0])
                class_split_len = int(class_len * split_ratio)
                print(class_len, class_split_len)
                test_images.extend(self.test_images[start:start+class_split_len])
                test_labels.extend(self.test_labels[start:start+class_split_len])
                start += class_len
            
            print(len(train_images), len(train_labels), len(test_images), len(test_labels))
            self.train_images, self.train_labels = np.array(train_images), np.array(train_labels)
            self.test_images, self.test_labels = np.array(test_images), np.array(test_labels)
            self.remaining_images, self.remaining_labels = np.array(remaining_images), np.array(remaining_labels)
                
            ##################################################
            
        elif dataset_name.lower() == "mstar":
#             train_dir = self.data_dir + '/Adj/17_DEG/'
#             test_dir = self.data_dir + '/Adj/15_DEG/'
            train_dir = self.data_dir + '/MSTAR-10/train/'
            test_dir = self.data_dir + '/MSTAR-10/test/'
            self.train_images, self.train_labels = self.get_mstar_data(train_dir)
            self.test_images, self.test_labels = self.get_mstar_data(test_dir)

#             self.train_images, self.train_labels = self.get_mstar_data_all_test(batch_filename='data_batch_1')
#             self.test_images, self.test_labels = self.get_mstar_data_all_test(batch_filename='test_batch')
            
            #### Split dataset into labeled and unlabeled ####
            self.train_images = np.array([x[1] for x in sorted(zip(self.train_labels, self.train_images), key=lambda tup: tup[0])])
            self.train_labels = np.array([x[0] for x in sorted(zip(self.train_labels, self.train_images), key=lambda tup: tup[0])])
            self.test_images = np.array([x[1] for x in sorted(zip(self.test_labels, self.test_images), key=lambda tup: tup[0])])
            self.test_labels = np.array([x[0] for x in sorted(zip(self.test_labels, self.test_images), key=lambda tup: tup[0])])
            
            remaining_images, remaining_labels = [], [] 
            remaining_images.extend(self.test_images)
            remaining_labels.extend(self.test_labels)
            
            print(len(self.train_images), len(self.train_labels), len(self.test_images), len(self.test_labels))
            self.remaining_images, self.remaining_labels = np.array(remaining_images), np.array(remaining_labels)
                
            ##################################################
            
            ##################################################
            
        elif dataset_name.lower() == "umist":
            train_dir = self.data_dir + '/face/'
            self.train_images, self.train_labels, self.test_images, self.test_labels = self.get_umist_data(train_dir)
            
            #### Split dataset into labeled and unlabeled ####
            self.train_images = np.array([x[1] for x in sorted(zip(self.train_labels, self.train_images), key=lambda tup: tup[0])])
            self.train_labels = np.array([x[0] for x in sorted(zip(self.train_labels, self.train_images), key=lambda tup: tup[0])])
            self.test_images = np.array([x[1] for x in sorted(zip(self.test_labels, self.test_images), key=lambda tup: tup[0])])
            self.test_labels = np.array([x[0] for x in sorted(zip(self.test_labels, self.test_images), key=lambda tup: tup[0])])
            
            remaining_images, remaining_labels = [], [] 
            remaining_images.extend(self.test_images)
            remaining_labels.extend(self.test_labels)
            
            print(len(self.train_images), len(self.train_labels), len(self.test_images), len(self.test_labels))
            self.remaining_images, self.remaining_labels = np.array(remaining_images), np.array(remaining_labels)
                
            ##################################################
        
        elif dataset_name.lower() == "breast":
            self.train_images, self.train_labels, self.test_images, self.test_labels = self.get_breast_data(train_split=0.7, scale=True)
            
            #### Split dataset into labeled and unlabeled ####
            self.train_images = np.array([x[1] for x in sorted(zip(self.train_labels, self.train_images), key=lambda tup: tup[0])])
            self.train_labels = np.array([x[0] for x in sorted(zip(self.train_labels, self.train_images), key=lambda tup: tup[0])])
            self.test_images = np.array([x[1] for x in sorted(zip(self.test_labels, self.test_images), key=lambda tup: tup[0])])
            self.test_labels = np.array([x[0] for x in sorted(zip(self.test_labels, self.test_images), key=lambda tup: tup[0])])
            
            remaining_images, remaining_labels = [], [] 
            remaining_images.extend(self.test_images)
            remaining_labels.extend(self.test_labels)
            
            print(len(self.train_images), len(self.train_labels), len(self.test_images), len(self.test_labels))
            self.remaining_images, self.remaining_labels = np.array(remaining_images), np.array(remaining_labels)
                
            ##################################################
        print("--------------------------------------------------------------------------")
        
        
    def get_breast_data(self, train_split=0.7, scale=False):
        self.class_names = [0, 1]
        series = pd.read_csv(self.data_dir + "breast.data", header=None, skiprows=None, index_col=0)
        print(series)
            
        values = series.values[:, 1:]
        
        if scale:
            scaler = MinMaxScaler()
            values = scaler.fit_transform(values)
            print("Scaled: ", values)
        
        labels = series.values[:, 0]
        
#         dataset_size = len(values)
#         train_size = int(dataset_size * train_split)
#         test_size = dataset_size - train_size
#         print(dataset_size, train_size, test_size)

#         train_images = values[:train_size]
#         test_images = values [train_size:]
#         train_labels = labels[:train_size]
#         test_labels = labels[train_size:]

        
        class_1_inds = np.where(labels == 0)[0]
        class_2_inds = np.where(labels == 1)[0]

        train_split = 0.7
        train_1_size = int(round(len(class_1_inds) * train_split, 0))
        test_1_size = len(class_1_inds) - train_1_size
        train_2_size = int(round(len(class_2_inds) * train_split, 0))
        test_2_size = len(class_2_inds) - train_2_size

        print(train_1_size, test_1_size, train_2_size, test_2_size)

        class_1_values = values[class_1_inds]
        class_1_labels = labels[class_1_inds]
        class_2_values = values[class_2_inds]
        class_2_labels = labels[class_2_inds]

        train_images = np.append(class_1_values[:train_1_size], class_2_values[:train_2_size], axis=0)
        test_images = np.append(class_1_values[train_1_size:], class_2_values[train_2_size:], axis=0)
        train_labels = np.append(class_1_labels[:train_1_size], class_2_labels[:train_2_size], axis=0)
        test_labels = np.append(class_1_labels[train_1_size:], class_2_labels[train_2_size:], axis=0)
        print(len(train_images), train_images)
        print(len(test_images), test_images)
        print(len(train_labels), train_labels)
        print(len(test_labels), test_labels)
    
        return train_images, train_labels, test_images, test_labels
    
    
    def get_mstar_data(self, data_dir, width=128, height=128, crop_size=128):
        dir_names = ["BMP2", "BTR70", "T72"] # 3-class
        self.class_names = dir_names
        X = []
        y = []

        for i in range(len(dir_names)):
            tmp_dir = data_dir + dir_names[i] + "/"
            sub_dirs = [x for x in os.listdir(tmp_dir)]
            for sub_dir in sub_dirs:
                tmp_dir = data_dir + dir_names[i] + "/" + sub_dir + "/"
                img_files = [x for x in os.listdir(tmp_dir) if x.endswith(".jpeg")]
                print(dir_names[i] + "/" + sub_dir, len(img_files))
                y += [i] * len(img_files)
                for j in range(len(img_files)):
                    img = im.imresize(im.imread((tmp_dir + img_files[j])), [height, width])
                    img = img[(height - crop_size) // 2 : height - (height - crop_size) // 2, \
                          (width - crop_size) // 2: width - (width - crop_size) // 2]
                    X.append(img)

        X_np = np.array(X)
#         X_np = X_np.reshape(X_np.shape[0], -1)
#         min_max_scaler = MinMaxScaler()
#         min_max_scaler.fit(X_np)
#         train_data_scaled = min_max_scaler.transform(X_np)
        
        return X_np, np.asarray(y)
    
    
    def get_mstar_data_all_test(self, batch_filename, scale=False):
        self.class_names = ["BMP2", "T72", "BTR70"]
        output_dir = self.data_dir + '/output/'
        with open(output_dir + batch_filename, 'rb') as f:
            mstar_dic = pickle.load(f)
        
        if scale:
#             scaler = MinMaxScaler(feature_range=(0, 1))
            scaler = StandardScaler()
            mstar_dic['data'] = scaler.fit_transform(mstar_dic['data'])
        
        mstar_dic['data'] = mstar_dic['data'] * 255.
        mstar_dic['data'] = mstar_dic['data'].astype('float32')
        
        return np.array(mstar_dic['data']), np.array(mstar_dic['labels'])
        
    
    def get_umist_data(self, data_dir, width=28, height=28, crop_size=28, train_ratio=0.9):
        sub_dir = ["1" + chr(97+x) for x in range(20)]
        print(sub_dir)
        self.class_names = sub_dir
        X_train, X_test = [], []
        y_train, y_test = [], []

        for i in range(len(sub_dir)):
            tmp_dir = data_dir + sub_dir[i] + "/"
            img_files = [x for x in os.listdir(tmp_dir) if x.endswith(".pgm")]
            print(sub_dir[i], len(img_files))
            train_len = int(len(img_files) * train_ratio)
            test_len = len(img_files) - train_len
            print("All, Train, Test: ", len(img_files), train_len, test_len)
            y_train += [i] * train_len
            y_test += [i] * test_len
            for j in range(len(img_files)):
                img = im.imresize(im.imread((tmp_dir + img_files[j])), [height, width])
                img = img[(height - crop_size) // 2 : height - (height - crop_size) // 2, \
                      (width - crop_size) // 2: width - (width - crop_size) // 2]
                if j < train_len:
                    X_train.append(img)
                else:
                    X_test.append(img)

        return np.asarray(X_train), np.asarray(y_train), np.asarray(X_test), np.asarray(y_test)
    
    def filter_dataset(self, images, labels, labels_2):
        images_filtered, labels_filtered = [], []
        for i in range(len(self.class_names)):
            class_label_1, class_label_2 = i, i
            inds = np.where(labels == class_label_1)[0]
            inds_2 = np.where(labels_2 == class_label_2)[0]

            num_samples = min(len(inds), len(inds_2))
#             num_samples -= num_samples % 100
            print(i, len(inds), len(inds_2), num_samples)
            
            images_filtered.extend(images[inds][:num_samples])
            labels_filtered.extend(labels[inds][:num_samples])
#         shuffler = np.random.RandomState(seed=42).permutation(len(labels_filtered))
#         return np.array(images_filtered)[shuffler], np.array(labels_filtered)[shuffler]
        return np.array(images_filtered), np.array(labels_filtered)
    
    def preprocess_dataset(self, dataset_2, reverse=False, scale=True):
        self.train_images, self.train_labels = self.filter_dataset(self.train_images, self.train_labels, dataset_2.train_labels)
        print(self.train_images.shape, self.train_labels.shape)
        self.test_images, self.test_labels = self.filter_dataset(self.test_images, self.test_labels, dataset_2.test_labels)
        print(self.test_images.shape, self.test_labels.shape)
        print(self.train_images[0])
        if scale:
            x_train, x_test = self.train_images.astype('float32')/255., self.test_images.astype('float32')/255.
            self.remaining_images = self.remaining_images.astype('float32')/255.
        else:
            x_train, x_test = self.train_images.astype('float32'), self.test_images.astype('float32')
        self.x_train, self.x_test = x_train.reshape(x_train.shape[0], -1), x_test.reshape(x_test.shape[0], -1)
        self.x_train_list = self.x_train.tolist()
        
        print(self.x_train[0])
        
        print(self.x_train.shape, self.x_test.shape)